"""
Package containing spectral reference information for common minerals and calibration targets. Reference spectra are
located in the reference package while information on specific *absorbtion* features is stored in the *features* package.
"""

from .generate import randomSpectra, genImage