"""
Contains measured reflectance spectra for PVC targets. Modify these or add new .txt files to change/add reference
PVC targets. For example *MyPVCSpectra.txt* would become accessible as
*hylite.reference.spectra.MyPVCSpectra*.
"""