"""
Put .txt files for custom reference spectra in here ( each line should contain a wavelength, reflectance pair) and they
will be loaded automatically as global variables. For example *MySpectra.txt* would become accessible as
*hylite.reference.spectra.MySpectra*.
"""