"""
Contains measured reflectance spectra for Spectralon targets. Modify these or add new .txt files to change/add reference
Spectralon targets. For example *MySpeccySpectra.txt* would become accessible as
*hylite.reference.spectra.MySpeccySpectra*.
"""